import os
import uuid
import time
import datetime
import math
import numpy as np

import bpy
import bmesh
from mathutils import Matrix, Vector, Quaternion, Color

from .debug import log, debug_mode
from .ops import HVE_OT_mechanist_base
from .mechanist import HVEMechanist

    
def makeObjectHeadlightLeft(ob, scn):
    type = "HVE_HEADLIGHT_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_HEADLIGHT_LEFT \n    DEF HVE_LIGHT_HEADLIGHT_LEFT_Low Switch {USE LIGHT_WHITE_LO}\n    DEF HVE_LIGHT_HEADLIGHT_LEFT_High Switch {USE LIGHT_WHITE_HI}\n"
    addHVELightText(ob, scn, type, text)
      
def makeObjectHeadlightRight(ob, scn):
    type = "HVE_HEADLIGHT_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_HEADLIGHT_RIGHT \n    DEF HVE_LIGHT_HEADLIGHT_RIGHT_Low Switch {USE LIGHT_WHITE_LO}\n    DEF HVE_LIGHT_HEADLIGHT_RIGHT_High Switch {USE LIGHT_WHITE_HI}\n"
    addHVELightText(ob, scn, type, text)
    
def makeObjectReverseLeft(ob, scn):
    type = "HVE_REVERSE_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_REVERSE_LEFT \n    DEF HVE_LIGHT_BACKUPLIGHT_LEFT Switch {USE LIGHT_WHITE_ON}\n"
    addHVELightText(ob, scn, type, text)
            
def makeObjectReverseRight(ob, scn):
    type = "HVE_REVERSE_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_REVERSE_RIGHT \n    DEF HVE_LIGHT_BACKUPLIGHT_RIGHT Switch {USE LIGHT_WHITE_ON}\n"
    addHVELightText(ob, scn, type, text)   
   
def makeObjectFogLightLeft(ob, scn):
    type = "HVE_FOGLIGHT_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_FOGLIGHT_LEFT \n    DEF HVE_LIGHT_HEADLIGHT_LEFT_Fog Switch {USE LIGHT_WHITE_ON}\n"
    addHVELightText(ob, scn, type, text)
  
def makeObjectFogLightRight(ob, scn):
    type = "HVE_FOGLIGHT_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_FOGLIGHT_RIGHT \n    DEF HVE_LIGHT_HEADLIGHT_RIGHT_Fog Switch {USE LIGHT_WHITE_ON}\n"
    addHVELightText(ob, scn, type, text)

def makeObjectAmberTurnLeft(ob, scn):
    type = "HVE_AMBERTURN_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_AMBERTURN_LEFT \n    DEF HVE_LIGHT_RUNNINGLIGHT_FRONT_LEFT Switch {USE LIGHT_AMBER_LO}\n    DEF HVE_LIGHT_SIGNALLIGHT_FRONT_LEFT Switch {USE LIGHT_AMBER_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_FRONT_LEFT Switch {USE LIGHT_AMBER_HI}\n"
    addHVELightText(ob, scn, type, text)
  
def makeObjectAmberTurnRight(ob, scn):
    type = "HVE_AMBERTURN_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_AMBERTURN_RIGHT \n    DEF HVE_LIGHT_RUNNINGLIGHT_FRONT_RIGHT Switch {USE LIGHT_AMBER_LO}\n    DEF HVE_LIGHT_SIGNALLIGHT_FRONT_RIGHT Switch {USE LIGHT_AMBER_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_FRONT_RIGHT Switch {USE LIGHT_AMBER_HI}\n"
    addHVELightText(ob, scn, type, text)
  
def makeObjectAmberTailLeft(ob, scn):
    type = "HVE_AMBERTAIL_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_AMBERTAIL_LEFT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_LEFT Switch {USE LIGHT_AMBER_LO}\n    DEF HVE_LIGHT_SIGNALLIGHT_REAR_LEFT Switch {USE LIGHT_AMBER_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_REAR_LEFT Switch {USE LIGHT_AMBER_HI}\n"
    addHVELightText(ob, scn, type, text)
   
def makeObjectAmberTailRight(ob, scn):
    type = "HVE_AMBERTAIL_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_AMBERTAIL_RIGHT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_RIGHT Switch {USE LIGHT_AMBER_LO}\n    DEF HVE_LIGHT_SIGNALLIGHT_REAR_RIGHT Switch {USE LIGHT_AMBER_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_REAR_RIGHT Switch {USE LIGHT_AMBER_HI}\n"
    addHVELightText(ob, scn, type, text) 
     
def makeObjectBrakeTurnLeft(ob, scn):
    type = "HVE_BRAKETURN_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_BRAKETURN_LEFT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_LEFT Switch {USE LIGHT_RED_LO}\n    DEF HVE_LIGHT_BRAKELIGHT_REAR_LEFT Switch {USE LIGHT_RED_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_REAR_LEFT Switch {USE LIGHT_RED_HI}\n    DEF HVE_LIGHT_SIGNALLIGHT_REAR_LEFT Switch {USE LIGHT_RED_HI}\n"
    addHVELightText(ob, scn, type, text)
      
def makeObjectBrakeTurnRight(ob, scn):
    type = "HVE_BRAKETURN_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_BRAKETURN_RIGHT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_RIGHT Switch {USE LIGHT_RED_LO}\n    DEF HVE_LIGHT_BRAKELIGHT_REAR_RIGHT Switch {USE LIGHT_RED_HI}\n    DEF HVE_LIGHT_EMERGENCYFLASHERLIGHT_REAR_RIGHT Switch {USE LIGHT_RED_HI}\n    DEF HVE_LIGHT_SIGNALLIGHT_REAR_RIGHT Switch {USE LIGHT_RED_HI}\n"
    addHVELightText(ob, scn, type, text) 
       
def makeObjectBrakeLeft(ob, scn):
    type = "HVE_BRAKE_LEFT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_BRAKE_LEFT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_LEFT Switch {USE LIGHT_RED_LO}\n    DEF HVE_LIGHT_BRAKELIGHT_REAR_LEFT Switch {USE LIGHT_RED_HI}\n"
    addHVELightText(ob, scn, type, text)
      
def makeObjectBrakeRight(ob, scn):
    type = "HVE_BRAKE_RIGHT"
    makeObjectHVELight(ob, scn, type)
    text = "#HVE_BRAKE_RIGHT \n    DEF HVE_LIGHT_RUNNINGLIGHT_REAR_RIGHT Switch {USE LIGHT_RED_LO}\n    DEF HVE_LIGHT_BRAKELIGHT_REAR_RIGHT Switch {USE LIGHT_RED_HI}\n"
    addHVELightText(ob, scn, type, text) 
      
def makeObjectBrakeCenter(ob, scn):
    type = "HVE_BRAKE_CENTER"
    HVEMechanist.makeObjectHVELight(ob, scn, type)
    text = "#HVE_BRAKE_CENTER \n    DEF HVE_LIGHT_BRAKELIGHT_CENTER Switch {USE LIGHT_RED_HI}\n"
    addHVELightText(ob, scn, type, text) 
     
class HVE_OT_make_light(HVE_OT_mechanist_base):
    bl_idname = "hve_lights.make_light"
    bl_label = "Make into HVE light"
    bl_description = "Make into an HVE light"



classes = (
    HVE_OT_make_light,
)
